header{* The Monad class *}
theory Monad imports MonadAxms
begin

text {* Declares monad syntax *}

constdefs 
  bind' :: "'a M => 'b M => 'b M"          (infixl ">>" 5)
  "bind' p q == bind p (% x. q)"

text{* This sets up a Haskell-style "{...}" syntax. *}

nonterminals
  monseq

syntax
  "_monseq"  :: "monseq=> 'a M"                   ("{(_)}"    [5] 5)
  "_mongen"  :: "[idts, 'a M, monseq]=> monseq"   ("_<-(_);//_"  [10, 6, 5] 5)
  "_monexp"  :: "['a M, monseq]=> monseq"         ("_;//_"       [6, 5] 5)
  "_monexp0" :: "['a M] => monseq"                ("(_)"         5)

translations
  "_monseq(_mongen x a r)" => "a >>= (% x.(_monseq r))"
  "_monseq(_monexp a r)"   => "a >> (_monseq r)"
  "_monseq(_monexp0 a)"    => "a"
  (* Retranslation of bind/seq into sequence notation *)
  "_monseq(_mongen x p q)"    <= "p >>= (%x. q)"
  "_monseq(_monexp p q)"      <= "p >> q"
  (* Normalization macros *)
  "_monseq(_monexp p q)"      <= "_monseq (_monexp p (_monseq q))"
  "_monseq(_mongen x p q)"    <= "_monseq (_mongen x p (_monseq q))"



text {* 
  Kleisli-application. We need one variant for each arity, unfortunately.
  *}
constdefs 
  kapp :: "('a=> 'b M)=> 'a M=> 'b M"        
 "kapp f x ==  {x<- x; f x}"
  kapp2 :: "('a=> 'b => 'c M)=> 'a M=> 'b M=> 'c M"        
 "kapp2 f x y == {x<- x; y<- y; f x y}"
  kapp3 :: "('a=> 'b => 'c => 'd M)=> 'a M=> 'b M=> 'c M=> 'd M"
 "kapp3 f x y z == {x<- x; y<- y; z<- z; f x y z}"

text {* 
  This sets up a uniform Kleisli application, i.e. 
  @term{f $$ x $$ y $$ z} for ternary operations. (Only up to arity 
  3 at the moment, I am afraid.) 
  *} 

syntax
 "_kapp"  :: "'x => 'y M=> 'z M"  (infixl "$$" 5)

translations
  "_kapp(_kapp(_kapp f x) y) z" == "kapp3 f x y z" 
  "_kapp(_kapp f x) y" == "kapp2 f x y" 
  "_kapp f x"  ==  "kapp f x" 

section{* Derived monad laws. *}

text{* Setting up the simplifier. *}

declare mon_lunit [simp]
        mon_runit [simp]
        mon_assoc [simp]
         
text{* More flavours of unit *}

theorem mon_lunit2 [simp]: "({eta x; p}) = p"
  by (unfold bind'_def, simp)

theorem mon_runite [simp]: "({x<- p; eta x}) = p"
  by simp



theorem mon_runit2 [simp]: "({p; eta ()}) = p" 
  by (unfold bind'_def, simp)


text{* 
  We need associativity in all different flavours with @term{bind'} 
  instead of @term{bind'}.
  We also need a better way of naming these :-)
  All of these lemmas should enable us to always message monadic sequences
  into a right-associative normal form like.
  *}

lemma mon_assoc2 [simp]: "({({x<- s; t x}); u}) = ({x<- s; t x; u})"
  by (unfold bind'_def, simp)

lemma mon_assoc4 [simp]: "({x<-({s; t}); u x}) = ({s; x<- t; u x})"
  by (unfold bind'_def, simp)

lemma mon_assoc5 [simp]: "({({s; t}); u})= ({s; t; u})"
  by (unfold bind'_def, simp)


section{* Map and join *}

constdefs
 mapF  :: "('a => 'b) => 'a M => 'b M"
"mapF f m == ({a<-m; eta (f a)})"

 joinM :: "('a M) M => 'a M "
"joinM m == ({a<-m; a})"


lemma mapF1[simp] : " (mapF f) o eta = eta o f " 
by (rule ext, simp add: mapF_def)

lemmas mapF1'[simp] = mapF1[simplified mapF_def o_def]

lemma mapF2[simp] : " (mapF f) o (mapF g) = mapF (f o g) "
by (rule ext, simp add: mapF_def)

lemmas mapF2'[simp] = mapF2[simplified o_def mapF_def]

lemma mapF3[simp] : "mapF id = id"
by (rule ext, simp add: mapF_def)

lemma joinM1[simp] : " joinM o eta = id "
by (rule ext, simp add: joinM_def)

lemmas joinM1'[simp] = joinM1[simplified joinM_def o_def]

lemma mapF_joinM1[simp] : " joinM (mapF f m) = (m >>= f) "
  by(simp add: joinM_def mapF_def)

lemmas mapF_joinM1'[simp] = mapF_joinM1[simplified joinM_def mapF_def]

lemma mapF_joinM2[simp] : " joinM o (mapF eta) = id "
by(rule ext, simp)

lemma mapF_joinM[simp] : " joinM o mapF (mapF f) = (mapF f) o joinM "
by (rule ext, simp add: mapF_def joinM_def)

lemmas mapF_joinM'[simp] = mapF_joinM[simplified mapF_def joinM_def o_def]; 

section{* Lifting *}

text{*
  Corresponds to the free functor from the base into the Kleisli category.
  We need one lift for each arity, due to currying.
  These are designed to work well with Kleisli application, i.e. you can write
  things like @term{lift2 f $$ a $$ b}.
  *}

constdefs 
  lift :: "('a => 'b) => 'a => 'b M" 
 "lift f == eta o f"
  lift2 :: "('a => 'b=> 'c) => 'a => 'b => 'c M"
 "lift2 f == %a. lift (f a)"
  lift3 :: "('a => 'b=> 'c=> 'd) => 'a => 'b=> 'c=> 'd M"
 "lift3 f a b == lift (f a b)" 

lemma lift_eta:  "(lift f $$ (eta x)) = eta (f x)"
 by (unfold lift_def kapp_def, simp)

lemma lift_eta2: "(lift f $$ ({p; q})) = ({p; lift f $$ q})"
 by (unfold lift_def kapp_def, simp)

lemma lift2_eta: "(lift2 f $$ (eta x) $$ (eta y)) = (eta (f x y))"
 by (simp add: lift2_def kapp2_def, 
     simp add: lift_def kapp_def)

lemma lift3_eta: "(lift3 f $$ (eta x) $$ (eta y)$$ (eta z)) = (eta (f x y z))"
 by (simp add: lift3_def kapp3_def, 
     simp add: lift_def kapp_def)




end
