

signature CTYPE =
sig

 type CType

 exception CType_arity of string

 val Type_CType : string list -> Term.typ -> CType
 val CType_Type : CType -> Term.typ
 val CType_Type' : CType -> Term.typ option
 val map_CType : (Symtab.key -> CType option) -> CType -> CType 

 val CType_of_TC : (string * int) -> CType 
 val de_Abs : CType -> (string * CType)  (* 'Replaces' all leading abstractions by applications of CFrees, 
                                             which are equal to fst o de_Abs *) 
 
 val beta_eta_norm : CType -> CType
 val rank_of_TC : theory -> string -> int
 val freevars : CType -> string list
 val well_formed : CType -> bool

end


structure CType : CTYPE =
struct

exception CType_arity of string

datatype CType = CVar of (int * bool)
               | TC of (string * int)
               | CApp of (CType * CType)
               | CAbs of CType
               | CFree of (string * sort) 


   fun lift k (CVar (x, b)) = (if (x < k) then CVar (x, b)
                               else (CVar (x + 1, b)))
     | lift k (CApp (a, b)) = CApp (lift k a, lift k b)
     | lift k (CAbs a) = (CAbs (lift (k + 1) a))
     | lift k x = x

   fun n_lift n k (CVar (x, b)) = (if (x < k) then CVar (x, b)
                                   else (CVar (x + n, b)))
     | n_lift n k (CApp (a, b)) = CApp (n_lift n k a, n_lift n k b)
     | n_lift n k (CAbs a) = (CAbs (n_lift n (k + 1) a))
     | n_lift n k x = x

   fun dsubst (CVar(x, b)) (i, t) = (if (i = x) then t
                                     else (if (i < x) then CVar(x - 1, b)
                                           else CVar (x, b)))
     | dsubst (CApp (a, b)) (i, t) = CApp (dsubst a (i, t), dsubst b (i, t))
     | dsubst (CAbs a) (i, t) = CAbs (dsubst a (i + 1, lift 0 t))
     | dsubst x (i, t) = x

   fun occurs1 (CVar (x, b)) v = (v = x)
     | occurs1 (CApp (s, t)) v = ((occurs1 s v) orelse (occurs1 t v))
     | occurs1 (CAbs t) v = occurs1 t (v + 1)
     | occurs1 _ v = false 

   fun occurs_unbound d (CVar(x, b)) = (d <= x)
     | occurs_unbound d (CFree _) = false (* ??? *)
     | occurs_unbound d (TC _) = false
     | occurs_unbound d (CApp (s, t)) = (occurs_unbound d s) orelse (occurs_unbound d t)
     | occurs_unbound d (CAbs s) = occurs_unbound (d + 1) s


fun freevars tc =
 let fun fv v (CVar _) = v
       | fv v (CFree (x, _)) = (if (x mem_string v) then v else (x::v))
       | fv v (TC _) = v
       | fv v (CApp (s, t)) = (fv v s) union (fv v t)
       | fv v (CAbs s) = fv v s
 in
  fv [] tc
 end

fun well_formed tc = not (occurs_unbound 0 tc)


(* Arity of a type constructor *)
fun rank_of_TC thy tc = 
  (case (Aux.type_arity(thy, tc)) of
           NONE => error("Unexpected exception: logical type "^tc^" has no arity")
         | SOME (a, i) => i)



(* Arity of CType *)
fun c_arity (TC (f, a)) = SOME a
  | c_arity (CApp (t1, t2)) = (case (c_arity t1) of
                                NONE => NONE
                              | SOME x => (case x of
                                             0 => NONE
                                           | _ => SOME (x - 1)))
  | c_arity (CAbs t) = (case (c_arity t) of
                         NONE => NONE
                       | SOME x => SOME (x + 1))
  | c_arity _ = SOME 0


(* Expansions *)
fun n_CAbs 0 t = t
  | n_CAbs n t = CAbs(n_CAbs (n - 1) t)

fun n_CApp 0 t = t
  | n_CApp n t = n_CApp (n - 1) (CApp (t, CVar(n - 1, true)))



fun eta_expand d ct = (let val ct' = n_lift d 0 ct
                           val ct'' = n_CApp d ct'
                       in
                        n_CAbs d ct''
                       end)


fun beta_expand t [] = t
  | beta_expand t (ct::cts) = beta_expand (CApp (CAbs (lift 0 t), ct)) cts


(* Beta-eta normal form for the special case of Isabelle-types where type variables are first-order
   i.e. it cannot occur an application of CVar to a term *)
fun beta (CAbs s) t = (dsubst s (0, t))
  | beta s t = CApp(s, t) 
fun eta (s as (CApp(t, CVar(0, _)))) =
     (if (occurs1 t 0) then (CAbs s)
      else (dsubst t (0, TC ("CType_arbitrary", 0))))
  | eta s = CAbs s
fun beta_eta_norm (CApp (s, t)) = beta (beta_eta_norm s) (beta_eta_norm t)
  | beta_eta_norm (CAbs s) = eta (beta_eta_norm s)
  | beta_eta_norm s = s







fun equalize_TC_arities (f, a) (g, a') = 
 (if (a <= a') then (g, n_CAbs (a' - a) (TC (f, a)))
  else (f, n_CAbs (a - a') (TC (g, a')))) 




(* mapping Type ---> CType, using a list of vars to be abstracted *)

fun Type_CType bvars T =  
 let fun index_ c v [] = NONE
       | index_ c v (x::xs) = (if (v = x) then (SOME c) else index_ (c + 1) v xs)
     fun index v = index_ 0 v (rev bvars)

     fun to_CType (TFree (v, s)) = (case (index v) of
                                           NONE =>  CFree (v, s)
                                         | SOME i => CVar (i, true))
       | to_CType (TVar _) = error ("Unexpected TVar in to_CType")
       | to_CType (Type (f, ts)) = 
        (let val r = length ts
             val tc = TC (f, r)
             val cts = map to_CType ts
             fun mk_app t [] = t
               | mk_app t (x::xs) = mk_app (CApp (t, x)) xs
         in
          mk_app tc cts
         end)
 in
   n_CAbs (length bvars) (to_CType (Aux.unvarifyT T))
 end



val CType_Type =
 let fun take_N 0 _ = []
       | take_N _ [] = []
       | take_N c (x::xs) = x::(take_N (c - 1) xs)

     fun from_CType buf (CFree(x, s)) = TFree(x, s)
       | from_CType buf (TC (f, r)) = (if (r <= length buf) then Type (f, take_N r buf)
                                       else raise CType_arity ("Arity error: TC " ^ f))
       | from_CType buf (CApp (a, b)) = (let val t = from_CType [] b
                                         in
                                          from_CType (t::buf) a
                                         end)
       | from_CType buf (CAbs a) = (if (occurs1 a 0) then raise CType_arity "Arity error: CAbs"
                                    else (from_CType buf a))
       | from_CType buf (CVar v) = raise CType_arity "Arity error: CVar"

  in
   (from_CType []) o beta_eta_norm
  end
 
fun CType_Type' x = SOME (CType_Type x)
                    handle (CType_arity s) => NONE 



(* Based on the assumption that (mp x) is a _closed_ CType for all x, if defined, i.e. we can skip lifting *)
fun map_CType mp ct =
 let fun map_CType_ mp (CApp(s, t)) = CApp(map_CType_ mp s, map_CType_ mp t)
       | map_CType_ mp (CAbs s) = CAbs (map_CType_ mp s)
       | map_CType_ mp (TC(f, a)) = (case (mp f) of
                                       SOME ct => ct
                                     | NONE => (TC(f, a)))
       | map_CType_ mp t = t
 in      
  map_CType_ mp ct
 end


fun CType_of_TC (tc, a) = eta_expand a (TC (tc, a))
  


fun de_Abs ct =
 let fun count_Abs (CAbs s) = 1 + (count_Abs s)
       | count_Abs _ = 0
     fun gen_args 0 = []
       | gen_args n = ("'" ^ (chr (96 + n)))::(gen_args (n-1))
     fun mk_App ct [] = ct
       | mk_App ct (v::vs) = mk_App (CApp(ct, CFree(v, ["HOL.type"]))) vs

     val c = count_Abs ct
     val args = gen_args c
 in
  if (1 < c) then ("(" ^ (commas args) ^ ")", mk_App ct args)
  else (commas args, mk_App ct args)
 end


end