(* **************************************************************************
 *
 * Inspecting and checking proof terms.
 *
 * $Source: /home/uniform/awe/repository/src/Extensions/inspect_pterms.sml,v $
 * $Revision: 755 $ of $Date: 2007-02-19 12:24:38 +0100 (Mo, 19 Feb 2007) $
 *)

signature INSPECT = 
  sig
      val get_ops_from_term   : term* (string* typ) list-> (string* typ) list
      val get_types_from_term : term* typ list-> typ list

      (* find all theorems, operations, types occuring in proof object *)
      val all_dep_thms_prf : (string * Proofterm.proof) -> string list
      val all_dep_thms     : thm-> string list
      val all_dep_ops      : thm-> (string* typ) list
      val all_dep_types    : thm-> typ list

      (* find only those not occuring to argument theory
       * or one of its ancestors *)
      val dep_thms_not_in  : theory-> thm-> string list
      val dep_ops_not_in   : theory-> thm-> (string*typ) list
      val dep_types_not_in : theory-> thm-> typ list

      (* find only those occuring in a list of theories *)
      val dep_thms_in_prf : theory list-> (string * Proofterm.proof) -> string list
      val dep_thms_in  : theory list-> thm-> string list
      val dep_ops_in   : theory list-> thm-> (string* typ) list
      val dep_types_in : theory list-> thm-> typ list 

      (* check wether we can abstract over theorem/opn/type/thy *)
      val check_abs_op : string -> thm -> string option
      val check_abs_thm : string -> thm -> string option
      val check_abs_type : typ -> thm -> string option

  end

structure Inspect : INSPECT = 
struct

local open Aux in 

(*
    fun lookup_type thy typnm = 
	let val tc= #tycons(Type.rep_tsig (Sign.tsig_of (sign_of thy)))
	in  Symtab.lookup(tc, typnm) 
	end
*)  


    (* -- Theorems ------------------------------------------------- *)

    fun addnew(xs, x) = x ins xs
(*  val full_proof_of = Aux.fullprf_of *)

    (* Find all theorems on which this depends, by name.*)
    fun all_dep_thms_prf (name, prf) =
	fold_thms_in_prf 
	(fn (nms, PThm((nm, _), prf', prop, _)) => (if (nm = name) 
                                                    then (all_dep_thms_prf (name, prf'))
                                                    else (nm ins nms))
	  | (nms, PAxm(nm, prop, _)) => nm ins nms)
	([], prf)

    val all_dep_thms = all_dep_thms_prf o Aux.fullprf_of'

    (* -- Operations ------------------------------------------------ *)

(*
    fun get_ops_from_term (t, ts)= 
	foldl_aterms (fn (ts, Const(tm, ty)) => addnew(ts, (tm, ty))
		       | (ts, _)             => ts) (ts, t)
*)
(* Isabelle2005 *)
    fun get_ops_from_term (t, ts)= 
	fold_aterms (fn (Const(tm, ty)) => (fn ts => addnew(ts, (tm, ty)))
		      | _ => I) t ts
    


    (* Get all operations used in a proof term *)
    fun all_dep_ops thm =
        fold_terms_in_prf get_ops_from_term ([], Aux.fullprf_of thm)
(* old:	Proofterm.fold_proof_terms
	  get_ops_from_term (fn (_, s)=> s) 
	  ([], Aux.fullprf_of thm) *)

    (* -- Types ---------------------------------------------------- *)

    fun get_types_from_term (t, ts) = fold_types (curry op ins) t ts
 
    (* Meaningful types-- one we can potentially abstract over. 
     * Builtin types like fun, prop, bool and dummy are not meaningful.
     * Free or schematic type variables are not meaningful.
     * Type names without a dot maybe meaningful (e.g. nat!)
     *)
    fun meaningful (Type (f, _)) = 
	not (f mem ["fun", "prop", "bool", "dummy"])
      | meaningful _             = false

    (* Get all types used in proof *)
    fun all_dep_types thm = 
	filter meaningful
	(Proofterm.fold_proof_terms 
	 get_types_from_term (op ins)
	 ([], Aux.fullprf_of thm))

    (* --- Global dep's - thms/ops/types not in this theory --------- *)

    (* Return only those identifiers which are "not_in" a certain theory,
     * i.e. those identifiers starting with <thyname>. where thyname is
     * not thy or one of thy's ancestors.
     * We also drop all unqualified identifiers, i.e. those with a dot;
     * on the assumption that these are types and operations from HOL.thy 
     * or below (e.g. ==> or prop).
     *
     * Question: Is this sufficient? Can we rely on fully qualified names?
     *)
    fun only_not_in thy =
	let val thynames = map dot_name (thy :: (ancestors_of thy))
	in  fn nm=> has_dot nm andalso
	            forall (fn th=> not (String.isPrefix th nm)) thynames
	end

    fun dep_thms_not_in thy thm = 
	filter (only_not_in thy) (all_dep_thms thm)


    (* Check wether an operation is not in a theory, by looking it up
     * in the signature. *)
    fun dep_ops_not_in thy thm  = 
	filter (is_none o Sign.const_type (sign_of thy) o fst)
	       (all_dep_ops thm)

    (* Check wether a type is not in a certain theory.
     * We can't trust fully qualified names here since e.g. nat or * 
     * in HOL are not fully qualified, so we look it up in the signature.
     *)
    fun dep_types_not_in thy thm = 
	filter (fn Type(nm, _) => (not(Sign.is_logtype thy nm)) (* is_none (lookup_type thy nm) *)
		 | _           => false) (all_dep_types thm)

(* old: try to certify.
	let val sg= sign_of thy
	in filter (fn tp as Type(nm, _)=> (ctyp_of sg tp; false) 
		                           handle (TYPE _) => true
		    | _                => false) (all_dep_types thm)
	end *)

    (* --- Local dep's - thms/ops/types from this theory ------------ *)

    fun only_in thys =
	let val thynames = map dot_name thys
	in  fn nm=> exists (fn th=> (String.isPrefix th nm)) thynames
	end

    fun dep_thms_in_prf thys = filter (only_in thys) o all_dep_thms_prf
    fun dep_thms_in thys = dep_thms_in_prf thys o Aux.fullprf_of'

    (* This is not very good-- we should look up operations and types
     * in the signature, just like above. The problem is that the
     * signature contains all ops and types from its ancestors as well,
     * so it's hard to tell if an op/types comes from _exactly_ here... 
     *)

    fun dep_ops_in thys thm   = filter ((only_in thys) o fst) (all_dep_ops thm)

    fun dep_types_in thys thm = 
	filter (fn Type(nm, _)=> only_in thys nm
                 | _          => false) (all_dep_types thm)

    (* --- Check functions ------------------------------------------------- 
     *
     * These functions check necessary but necessarily sufficient conditions
     * for the succes of the abstract functions.
     * i.e. if these functions return an error (SOME errormessage), then the
     * abstraction function will fail, but if they return no error (NONE),
     * the abstraction may still not suceed.
     *)

    (* Check wether we can abstraction from a certain theorem. *)
    (* Necessary conditions:
     * - theorem occurs somewhere in proof term (ows. abstraction is vacuous) 
     *)

    fun check_abs_thm thmnm thm =
	if not (thmnm mem all_dep_thms thm) then
	    SOME ("Theorem "^thmnm^" does not occur in proof at all")
        else NONE

    (* Check wether we can abstract from a certain operation.
     * Necessary conditions:
     * - operation does not occur in the statement of a theorem used in 
     *   the proof.
     * - operation occurs somewhere in proof term
     *)
    fun check_abs_op opnm thm =
	let val prfsmbs = fold_thms_in_prf 
	    (fn (ls, PThm(_, _, prop, _))=> 
	     ls union (get_ops_from_term (prop ,[]))
	  | (ls, PAxm(_, prop, _))=> 
		ls union (get_ops_from_term (prop, []))) 
	    ([], Aux.fullprf_of thm)
	    val thmsmbs = all_dep_ops thm
	in  if opnm mem (map fst prfsmbs) then
	    SOME ("Operation "^opnm^" occurs in theorems used in proof")
	    else if not (opnm mem (map fst thmsmbs)) then
		  SOME ("Operation "^opnm^" does not occur in proof at all")
		 else NONE
	end

    (* Check we can abstract over a type:
     * - type does not occur in theorems used in proof
     * - type does occur somewhere in the proof at all
     *
     * Turns it this is too liberal, we need to check wether type
     * occurs in operation used in assumptions-- or something.
     *)
    fun get_types_from_ops prf =
	Proofterm.fold_proof_terms 
	  get_types_from_term (op ins)
	  ([], prf)
(*
    fun get_classes_from_typ t =
	foldl_atyps (fn (more, TFree (_, cs))=> cs union more
		      | (more, TVar (_, cs)) => cs union more) ([], t)
*)
    fun get_classes_from_typ t =
	fold_atyps (fn (TFree (_, cs)) => (fn more => (cs union more))
		     | (TVar (_, cs)) => (fn more => (cs union more))) t []

    fun check_abs_type typ thm =
	let val prftyps = fold_thms_in_prf
	    (fn (ts, PThm(_, _, prop, _))=> 
		         get_types_from_term(prop, []) union ts
	      | (ts, PAxm(_, prop, _))=> 
		         get_types_from_term(prop, []) union ts
	      | (ts, _) => ts)
            ([], Aux.fullprf_of thm) 
(*	    val optyps  = get_types_from_ops (Aux.fullprf_of thm) *)
	    val tpstr   = Sign.string_of_typ (sign_of_thm thm) typ
	in  if (typ mem prftyps) then 
	    SOME ("Type "^tpstr^" occurs in theorems used in proof.")
	    else if not (typ mem (all_dep_types thm)) then
		SOME ("Type "^tpstr^" does not occur in proof at all.")
		 else NONE
	end

end (* local open Aux *)
end
